import contextlib
import signal
from asyncio import CancelledError, Task
from typing import Any, override

from greyhorse.app.abc.services import ServiceError, ServiceState
from greyhorse.app.entities.services import AsyncService
from greyhorse.app.resources import inject
from greyhorse.result import Result
from greyhorse.utils.invoke import get_asyncio_loop
from uvicorn import Config, Server

from greyhorse_web.common import ASGIApp


class UvicornService(AsyncService):
    def __init__(self, host: str, port: int, config: dict[str, Any] | None = None) -> None:
        super().__init__()
        self._host = host
        self._port = port
        self._conf = config or {}
        self._server: Server | None = None
        self._task: Task | None = None

    @override
    @inject
    async def setup(self, app: ASGIApp | None) -> Result[ServiceState, ServiceError]:
        if app is None:
            return ServiceError.NoSuchResource(name='Application for Uvicorn').to_result()

        config = Config(app, host=self._host, port=self._port, **self._conf)
        self._server = Server(config)
        return await super().setup()

    @override
    async def teardown(self) -> Result[ServiceState, ServiceError]:
        self._server = None
        return await super().teardown()

    async def start(self) -> None:
        if self._task is not None:
            return

        self._task = get_asyncio_loop().create_task(self._server.serve())
        await self._switch_to_active(True)

    async def stop(self) -> None:
        if self._task is None:
            return

        self._server.handle_exit(signal.SIGINT, None)

        with contextlib.suppress(CancelledError):
            await self._task

        self._task = None
        await self._switch_to_active(False)



from starlette.requests import Request
from starlette.websockets import WebSocket

from greyhorse_web.schemas import ClientInfo


def get_client_info(request: Request | WebSocket) -> ClientInfo:
    client_ip = request.client.host.strip()
    client_port = request.client.port
    user_agent = request.headers.get('user-agent')
    referrer = request.headers.get('referer')

    x_real_ip = request.headers.get('x-real-ip', '').strip()
    x_fwd_ip = request.headers.get('x-forwarded-for', '').split(',')
    if isinstance(x_fwd_ip, tuple | list) and len(x_fwd_ip) > 0:
        x_fwd_ip = x_fwd_ip[0].strip()
    ip_current = x_fwd_ip or x_real_ip or client_ip

    return ClientInfo(
        ip=ip_current,
        port=client_port,
        host=request.base_url.netloc,
        agent=user_agent,
        referrer=referrer,
        scheme=request.base_url.scheme,
    )



from typing import override

from fastapi import APIRouter, FastAPI
from greyhorse.app.abc.controllers import ControllerError
from greyhorse.app.entities.controllers import ResConf, ResourceController
from greyhorse.app.registries import MutDictRegistry
from greyhorse.app.resources import Container
from greyhorse.result import Ok, Result

from . import FastAPIRouterCollector


class FastAPIController(ResourceController):
    def __init__(self, prefix: str = '') -> None:
        super().__init__(
            resources=[
                ResConf(type=FastAPIRouterCollector, required=False),
                ResConf(type=FastAPI, required=False),
            ]
        )
        self._prefix = prefix
        self._router: FastAPI | APIRouter | None = None
        self._parent: FastAPIRouterCollector | None = None
        self._children = MutDictRegistry[str, APIRouter](allow_many=True)

    @override
    def setup(self, container: Container) -> Result[bool, ControllerError]:
        if not (res := super().setup(container)):
            return res

        if (parent := container.get(FastAPIRouterCollector).unwrap_or_none()) is not None:
            self._parent = parent
        elif app := container.get(FastAPI).unwrap_or_none():
            self._router = app
        else:
            return ControllerError.NoSuchResource(
                name='FastAPIRouterCollector/FastAPI'
            ).to_result()

        if self._router is None:
            self._router = APIRouter(prefix=self._prefix)
        if self._parent is not None:
            self._parent.add(self._prefix, self._router)

        res = True
        if self._parent is not None:
            res &= container.registry.remove_factory(FastAPIRouterCollector)
        res &= container.registry.add_factory(FastAPIRouterCollector, self._children)
        return Ok(res)

    @override
    def teardown(self, container: Container) -> Result[bool, ControllerError]:
        if self._parent is not None:
            self._parent.remove(self._prefix, self._router)

        container.registry.remove_factory(FastAPIRouterCollector)
        if self._parent is not None:
            container.registry.add_factory(FastAPIRouterCollector, self._parent)

        self._parent = self._router = None
        return super().teardown(container)

    def start(self) -> None:
        for path, handler, kwargs in self._children.list_with_metadata():
            self._router.include_router(handler, prefix=path, **kwargs)



import base64
import math
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, NewType

import strawberry
from greyhorse.data.query import Query
from greyhorse.data.repositories import AsyncFilterable
from pydantic import BaseModel, Field, model_validator
from strawberry import scalar
from strawberry.types import Info
from strawberry.types.nodes import Selection

from .context import Context
from .mixins import FromEntityMixin

Encoded = scalar(
    NewType('Encoded', bytes),
    description=('Represents encoded cursor data'),
    serialize=lambda v: base64.b85encode(v).decode('latin1'),
    parse_value=lambda v: base64.b85decode(v.encode('latin1')),
)


@strawberry.type
class PageInfo:
    has_next: bool
    has_prev: bool
    count: int
    total: int
    page: int
    page_size: int
    total_pages: int
    start_cursor: Encoded | None
    end_cursor: Encoded | None


@strawberry.type
class Edge[T]:
    node: T
    cursor: Encoded


@strawberry.type
class Connection[T]:
    page_info: PageInfo
    edges: list[Edge[T]]
    nodes: list[T]


DEFAULT_LIMIT_VALUE = 10
MAX_LIMIT_VALUE = 1000


def cursor_encode(ident: object) -> str:
    return base64.b85encode(str(ident).encode('latin1')).decode('latin1')


def cursor_decode(code: str) -> str:
    return base64.b85decode(code.encode('latin1')).decode('latin1')


@dataclass(slots=True, frozen=True)
class NeedFieldData:
    need_has_next: bool
    need_has_prev: bool
    need_total: bool
    need_page: bool


class RelayCursorInfo(BaseModel, frozen=True):
    first: int | None = Field(gt=0, le=MAX_LIMIT_VALUE, default=None)
    last: int | None = Field(gt=0, le=MAX_LIMIT_VALUE, default=None)
    after: bytes | None = None
    before: bytes | None = None

    @model_validator(mode='before')
    @classmethod
    def check(cls, values: dict[str, Any]) -> dict[str, Any]:
        if 'first' in values and 'last' in values:
            raise ValueError('You can specify only "first" or "last"')
        if 'after' in values and 'before' in values:
            raise ValueError('You can specify only "after" or "before"')
        if 'last' in values and 'after' in values:
            raise ValueError('You can specify "after" with only "first"')
        if 'first' in values and 'before' in values:
            raise ValueError('You can specify "before" with only "last"')
        if 'first' not in values and 'last' not in values:
            if 'first' not in values and 'after' in values:
                values['first'] = DEFAULT_LIMIT_VALUE
            elif 'last' not in values and 'before' in values:
                values['last'] = DEFAULT_LIMIT_VALUE
            else:
                values['first'] = DEFAULT_LIMIT_VALUE
        return values

    @classmethod
    def get(
        cls,
        first: int | None = None,
        last: int | None = None,
        after: Encoded | None = None,
        before: Encoded | None = None,
    ) -> 'RelayCursorInfo':
        data = dict()
        if first is not None:
            data['first'] = first
        if last is not None:
            data['last'] = last
        if after is not None:
            data['after'] = after
        if before is not None:
            data['before'] = before
        return cls(**data)

    def decoded_after(
        self, as_: Callable[[str], Any] | None = None, default: object | None = None
    ) -> object | None:
        if self.after is None:
            return default
        res = self.after.decode('latin1')
        if as_:
            return as_(res)
        return res

    def decoded_before(
        self, as_: Callable[[str], Any] | None = None, default: object | None = None
    ) -> object | None:
        if self.before is None:
            return default
        res = self.before.decode('latin1')
        if as_:
            return as_(res)
        return res

    @staticmethod
    def encode_cursor(value: object) -> bytes:
        return str(value).encode('latin1')

    @staticmethod
    def need_has_fields(info: Info[Context]) -> NeedFieldData:
        need_has_next = need_has_prev = need_total = need_page = False

        def traverse(selection: Selection) -> bool:
            nonlocal need_has_next, need_has_prev, need_total, need_page

            if selection.name == 'pageInfo':
                for subfield in selection.selections:
                    if subfield.name == 'hasNext':
                        need_has_next = True
                    elif subfield.name == 'hasPrev':
                        need_has_prev = True
                    elif subfield.name == 'total':
                        need_total = True
                    elif subfield.name == 'page':
                        need_page = True
                return True

            return any(traverse(subfield) for subfield in selection.selections)

        for field in info.selected_fields:
            if traverse(field):
                break

        return NeedFieldData(
            need_has_next=need_has_next,
            need_has_prev=need_has_prev,
            need_total=need_total,
            need_page=need_page,
        )


class AsyncPaginator[E, ID]:
    def __init__(
        self,
        out_model: type[FromEntityMixin],
        repo: AsyncFilterable[E, ID],
        after_filter: Callable[[ID], Any],
        before_filter: Callable[[ID], Any],
        id_fn: Callable[[E], ID],
        str_to_id: Callable[[str], ID] = int,
    ) -> None:
        self._out_model = out_model
        self._repo = repo
        self._after_filter = after_filter
        self._before_filter = before_filter
        self._id_fn = id_fn
        self._str_to_id = str_to_id

    async def __call__(
        self,
        info: Info[Context],
        query: Query,
        first: int | None = None,
        last: int | None = None,
        after: Encoded | None = None,
        before: Encoded | None = None,
    ) -> Connection:
        cur_info = RelayCursorInfo.get(first, last, after, before)
        after_id = cur_info.decoded_after(as_=self._str_to_id)
        before_id = cur_info.decoded_before(as_=self._str_to_id)

        need_data = RelayCursorInfo.need_has_fields(info)
        total = await self._repo.count(query) if need_data.need_total else 0

        if after:
            list_query = query.add_filter(self._after_filter(after_id))
        elif before:
            list_query = query.add_filter(self._before_filter(before_id))
        else:
            list_query = query.clone()

        count = cur_info.first or cur_info.last or DEFAULT_LIMIT_VALUE
        objects = list(await self._repo.list(list_query, limit=count))

        has_prev = has_next = False
        count_prev = 0

        if objects:
            if need_data.need_has_next:
                last_obj_id = self._id_fn(objects[-1])
                has_next = await self._repo.exists_by(
                    query.add_filter(self._after_filter(last_obj_id))
                )

            if need_data.need_page:
                first_obj_id = self._id_fn(objects[0])
                count_prev = await self._repo.count(
                    query.add_filter(self._before_filter(first_obj_id))
                )
                has_prev = count_prev > 0
            elif need_data.need_has_prev:
                first_obj_id = self._id_fn(objects[0])
                has_prev = await self._repo.exists_by(
                    query.add_filter(self._before_filter(first_obj_id))
                )

        edges = [
            Edge(
                cursor=RelayCursorInfo.encode_cursor(self._id_fn(instance)),
                node=self._out_model.from_entity(instance),
            )
            for instance in objects
        ]

        return Connection(
            page_info=PageInfo(
                has_next=has_next,
                has_prev=has_prev,
                count=len(edges),
                total=total,
                page=math.ceil(count_prev / count) + 1,
                page_size=count,
                total_pages=math.ceil(total / count),
                start_cursor=edges[0].cursor if edges else None,
                end_cursor=edges[-1].cursor if edges else None,
            ),
            edges=edges,
            nodes=[edge.node for edge in edges],
        )



from typing import override

from fastapi import HTTPException
from greyhorse.utils import json
from orjson import orjson
from strawberry.fastapi import GraphQLRouter
from strawberry.http import GraphQLHTTPResponse


class JsonGraphQLRouter(GraphQLRouter):
    use_indent: bool = False
    sort_keys: bool = False

    @override
    def parse_json(self, data: str | bytes) -> object:
        try:
            return json.loads(data)
        except orjson.JSONDecodeError as e:
            raise HTTPException(400, 'Unable to parse request body as JSON') from e

    @override
    def encode_json(self, data: GraphQLHTTPResponse) -> str:
        try:
            return json.dumps(data)
        except orjson.JSONEncodeError as e:
            raise HTTPException(400, 'Unable to encode response body as JSON') from e



from functools import cached_property

from greyhorse.app.registries import MutDictRegistry
from greyhorse.app.resources import Container
from greyhorse.maybe import Maybe
from strawberry.fastapi import BaseContext


class Context(BaseContext):
    __slots__ = ('_containers_map',)

    def __init__(self) -> None:
        super().__init__()
        self._containers_map = MutDictRegistry[type, Container]()

    def get[T](self, hint: type[T]) -> Maybe[T]:
        return self._request_container.get(hint)

    @cached_property
    def _request_container(self) -> Container:
        return self.request.state.container

    def container_for(self, source: object) -> Container:
        return self._containers_map.get(type(source)).unwrap_or_else(
            lambda: self._request_container
        )

    def add_source_container(self, source: object, container: Container) -> bool:
        return self._containers_map.add(type(source), container)

    def remove_source_container(self, source: object, container: Container) -> bool:
        return self._containers_map.remove(type(source), container)



from functools import partial
from typing import override

from fastapi import APIRouter, FastAPI
from greyhorse.app.abc.providers import SharedProvider
from greyhorse.app.abc.resources import Lifetime
from greyhorse.app.abc.services import ServiceError, ServiceState
from greyhorse.app.entities.services import AsyncService, provide
from greyhorse.app.registries import MutDictRegistry
from greyhorse.result import Result

from greyhorse_web.common import ASGIApp

from . import FastAPIRouterCollector
from .factory import create_fastapi


class FastAPIService(AsyncService):
    def __init__(
        self,
        title: str,
        debug: bool = False,
        root_path: str = '',
        version: str | None = None,
        cors: list[str] | None = None,
    ) -> None:
        super().__init__()
        self._factory = partial(create_fastapi, title, debug, root_path, version, cors)
        self._app: FastAPI | None = None
        self._handlers = MutDictRegistry[str, APIRouter](allow_many=True)

    @override
    async def setup(self) -> Result[ServiceState, ServiceError]:
        self._app = self._factory()
        return await super().setup()

    @override
    async def teardown(self) -> Result[ServiceState, ServiceError]:
        self._app = None
        return await super().teardown()

    async def start(self) -> None:
        for path, handler, kwargs in self._handlers.list_with_metadata():
            self._app.include_router(handler, prefix=path, **kwargs)
        await self._switch_to_active(True)

    async def stop(self) -> None:
        await self._switch_to_active(False)

    @provide(lifetime=Lifetime.COMPONENT())
    def create_collector_provider(self) -> SharedProvider[FastAPIRouterCollector]:
        yield self._handlers

    @provide(lifetime=Lifetime.COMPONENT())
    def create_asgi(self) -> ASGIApp:
        return self._app



import dataclasses
from collections.abc import Iterable
from typing import Any

import strawberry


class FromEntityMixin[E, T]:
    @classmethod
    def from_entity(cls, entity: E) -> T:
        data = {}
        for field in dataclasses.fields(cls):
            if type(field) is dataclasses.Field:
                data[field.name] = getattr(entity, field.name)
        return cls(**data)

    @classmethod
    def from_entities(cls, entities: Iterable[E]) -> Iterable[T]:
        for entity in entities:
            yield cls.from_entity(entity)


class ToDictMixin:
    def to_dict(self) -> dict[str, Any]:
        return {k: v for k, v in dataclasses.asdict(self).items() if v != strawberry.UNSET}



from pydantic import BaseModel, IPvAnyAddress


class ClientInfo(BaseModel, frozen=True):
    ip: IPvAnyAddress
    port: int
    host: str
    agent: str | None = None
    referrer: str | None = None
    scheme: str | None = None



import asyncio
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from functools import partial

import pydantic
from fastapi import FastAPI, Request, Response
from fastapi.exceptions import RequestValidationError
from fastapi.middleware.cors import CORSMiddleware
from greyhorse.app.entities.application import Application
from greyhorse.app.runtime import Runtime
from greyhorse.utils.json import dumps_raw, loads
from starlette.authentication import (
    AuthCredentials,
    AuthenticationBackend,
    BaseUser,
    UnauthenticatedUser,
)
from starlette.middleware.authentication import AuthenticationMiddleware


class ORJsonRequest(Request):
    async def json(self) -> object:
        if not hasattr(self, '_json'):
            body = await self.body()
            self._json = loads(body)
        return self._json


class ORJsonResponse(Response):
    media_type = 'application/json'
    use_indent: bool = False
    sort_keys: bool = False

    def render(self, content: object) -> bytes:
        return dumps_raw(content, use_indent=self.use_indent, sort_keys=self.sort_keys)


class AuthBackend(AuthenticationBackend):
    async def authenticate(self, _: Request) -> tuple[AuthCredentials, BaseUser] | None:
        return AuthCredentials(), UnauthenticatedUser()


@asynccontextmanager
async def fastapi_lifespan(_: FastAPI, app: Application) -> AbstractAsyncContextManager:
    runtime = Runtime()
    asyncio.set_event_loop(runtime.loop)

    with runtime:
        app.setup().unwrap()
        app.start()

        yield

        app.stop()
        app.teardown().unwrap()
        app.unload().unwrap()


def create_fastapi(
    title: str,
    debug: bool = False,
    root_path: str = '',
    version: str | None = None,
    cors: list[str] | None = None,
    application: Application | None = None,
) -> FastAPI:
    app = FastAPI(
        title=title,
        version=version,
        debug=debug,
        root_path=root_path,
        lifespan=partial(fastapi_lifespan, app=application) if application else None,
        default_response_class=ORJsonResponse,
    )

    app.add_middleware(
        CORSMiddleware,  # type: ignore
        allow_origins=[str(origin) for origin in cors] if cors else ['*'],
        allow_credentials=True,
        allow_methods=['*'],
        allow_headers=['*'],
    )

    app.add_middleware(
        AuthenticationMiddleware,  # type: ignore
        backend=AuthBackend(),
    )

    ORJsonResponse.use_indent = ORJsonResponse.sort_keys = debug
    Request.json = ORJsonRequest.json

    @app.exception_handler(RequestValidationError)
    async def validation_exception_handler(
        _: Request, exc: pydantic.ValidationError
    ) -> ORJsonResponse:
        error_dict = {}
        for e in exc.errors():  # type: dict[str, str]
            location = '.'.join([str(loc) for loc in e['loc'] if loc != 'body'])
            if location in error_dict:
                error_dict[location] = [
                    error_dict[location],
                    {'msg': e['msg'], 'type': e['type']},
                ]
            else:
                error_dict[location] = {'msg': e['msg'], 'type': e['type']}

        return ORJsonResponse(error_dict, status_code=422)

    return app



from datetime import timedelta

import strawberry
from pytimeparse.timeparse import timeparse

TimeDelta = strawberry.scalar(
    timedelta, serialize=lambda v: str(v), parse_value=lambda v: timedelta(seconds=timeparse(v))
)



from collections.abc import Awaitable, Callable
from typing import Any, override

from greyhorse.app.abc.resources import Lifetime
from greyhorse.app.resources import Container
from greyhorse.maybe import Nothing
from greyhorse.utils.types import (
    is_awaitable,
    is_maybe,
    is_optional,
    unwrap_maybe,
    unwrap_optional,
)
from strawberry import Info
from strawberry.extensions import FieldExtension, SchemaExtension
from strawberry.types.arguments import StrawberryArgument
from strawberry.types.base import StrawberryOptional
from strawberry.types.field import StrawberryField

from .context import Context


class ContainerExtension(SchemaExtension):
    def resolve(
        self,
        next_: Callable[..., object | Awaitable[object]],
        source: object,
        info: Info[Context],
        *args: object,
        **kwargs: dict[str, Any],
    ) -> object:
        if container := getattr(source, '__container__', None):
            if is_awaitable(next_):
                return self._wrapper(container, next_, source, info, *args, **kwargs)

            with container(lifetime=Lifetime.REQUEST()) as request_container:
                info.context.add_source_container(source, request_container)
                res = next_(source, info, *args, **kwargs)
                info.context.remove_source_container(source, request_container)
        else:
            res = next_(source, info, *args, **kwargs)
        return res

    @staticmethod
    async def _wrapper(
        container: Container,
        next_: Callable[..., Awaitable[object]],
        source: object,
        info: Info[Context],
        *args: object,
        **kwargs: dict[str, Any],
    ) -> object:
        with container(lifetime=Lifetime.REQUEST()) as request_container:
            info.context.add_source_container(source, request_container)
            res = await next_(source, info, *args, **kwargs)
            info.context.remove_source_container(source, request_container)
            return res


class Provide:
    pass


class Depends(FieldExtension):
    __slots__ = ('_args',)

    def __init__(self) -> None:
        self._args: list[StrawberryArgument] = []

    @override
    def apply(self, field: StrawberryField) -> None:
        keep_arguments = []

        for arg in field.arguments:
            if isinstance(arg.default, Provide):
                self._args.append(arg)
                continue
            keep_arguments.append(arg)

        field.arguments = keep_arguments

    @override
    def resolve(
        self,
        next_: Callable[..., Any],
        source: object,
        info: Info[Context],
        *args: object,
        **kwargs: dict[str, Any],
    ) -> Any:
        kw = self._inject_params(source, info, **kwargs)
        return next_(source, info, *args, **kw)

    @override
    async def resolve_async(
        self,
        next_: Callable[..., Any],
        source: object,
        info: Info[Context],
        *args: object,
        **kwargs: dict[str, Any],
    ) -> Any:
        kw = self._inject_params(source, info, **kwargs)
        return await next_(source, info, *args, **kw)

    def _inject_params(
        self, source: object, info: Info[Context], **kwargs: dict[str, Any]
    ) -> dict[str, Any]:
        container = info.context.container_for(source)
        kw = {}

        for arg in self._args:
            if isinstance(arg.type, StrawberryOptional):
                arg_type = arg.type.of_type | None
            else:
                arg_type = arg.type

            if (value := container.get(unwrap_maybe(unwrap_optional(arg_type)))) or (
                value := info.context.get(unwrap_maybe(unwrap_optional(arg_type)))
            ):
                kw[arg.python_name] = value if is_maybe(arg_type) else value.unwrap()
            elif is_maybe(arg_type):
                kw[arg.python_name] = Nothing
            elif is_optional(arg_type):
                kw[arg.python_name] = None
            kw.update(kwargs)

        return kw



import contextlib
from asyncio import CancelledError, Task

from greyhorse.app.abc.collectors import Collector
from greyhorse.app.abc.providers import SharedProvider
from greyhorse.app.abc.resources import Lifetime
from greyhorse.app.entities.services import AsyncService, provide
from greyhorse.app.registries import DictRegistry
from greyhorse.utils.invoke import get_asyncio_loop
from grpclib.server import Server

from greyhorse_web.common import GRPCHandler

GRPCHandlerCollector = Collector[str, GRPCHandler]


class GRPCService(AsyncService):
    def __init__(self, host: str, port: int) -> None:
        super().__init__()
        self._host = host
        self._port = port
        self._server: Server | None = None
        self._task: Task | None = None
        self._handlers = DictRegistry[str, GRPCHandler](allow_many=True)

    @provide(lifetime=Lifetime.COMPONENT())
    def create_collector(self) -> SharedProvider[GRPCHandlerCollector]:
        yield self._handlers

    async def start(self) -> None:
        if self._task is not None:
            return

        self._server = Server([h[1] for h in self._handlers.list()])
        await self._server.start(self._host, self._port)
        self._task = get_asyncio_loop().create_task(self._server.wait_closed())
        await self._switch_to_active(True)

    async def stop(self) -> None:
        if self._task is None:
            return

        self._server.close()
        self._task.cancel()

        with contextlib.suppress(CancelledError):
            await self._task

        self._task = None
        self._server = None
        await self._switch_to_active(False)



from collections.abc import AsyncGenerator

from fastapi import Request, WebSocket, params
from greyhorse.app.abc.resources import Lifetime
from greyhorse.app.resources import Container
from greyhorse.maybe import Maybe

from greyhorse_web.schemas import ClientInfo

from .utils import get_client_info


class Provide[T](params.Depends):
    def __init__(self, hint: type[T], optional: bool = False) -> None:
        super().__init__(dependency=self.__call__, use_cache=True)
        self._hint = hint
        self._optional = optional

    async def __call__(self, request: Request) -> Maybe[T] | T:
        value = request.state.container.get(self._hint)
        return value if self._optional else value.expect(f'Cannot unwrap "{self._hint}"')


async def request_container(
    container: Container, request: Request
) -> AsyncGenerator[None, None]:
    context = {Request: request, ClientInfo: get_client_info(request)}
    lifetime = Lifetime.REQUEST()

    with container(context, lifetime) as request_container:
        request.state.container = request_container
        yield


async def websocket_container(
    container: Container, request: WebSocket
) -> AsyncGenerator[None, None]:
    context = {WebSocket: request, ClientInfo: get_client_info(request)}
    lifetime = Lifetime.SESSION()

    with container(context, lifetime) as request_container:
        request.state.container = request_container
        yield
